import logging
import typing


__all__ = ["Dropped", "ItemGetValueError", "ExceptionFilter"]


class Dropped(Exception):
    """Raise when pipeline dropped one object"""


class ItemGetValueError(Exception):
    """Raise when get value by wrong key"""


class ExceptionFilter(logging.Filter):
    """A exception log filter class for logging."""

    def __init__(
        self,
        exceptions: typing.Sequence[typing.Type[Exception]] = (Dropped,),
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.exceptions = exceptions

    def filter(self, record: logging.LogRecord) -> bool:
        if record.exc_info:
            for e in self.exceptions:
                if record.exc_info[0] is e:
                    return False
        return True



import typing
import asyncio
import tempfile
import os
import webbrowser
from contextlib import contextmanager
from logging import Logger

from tenacity import retry as _retry
from tenacity.wait import wait_fixed
from tenacity.stop import stop_after_attempt


def retry(
    retries: int, delay: float
) -> typing.Callable[[typing.Callable], typing.Callable]:
    return _retry(
        wait=wait_fixed(delay), stop=stop_after_attempt(retries + 1), reraise=True
    )


async def run_cor_func(func: typing.Callable, *args, **kwargs) -> typing.Any:
    ret = func(*args, **kwargs)
    if asyncio.iscoroutine(ret):
        ret = await ret

    return ret


@contextmanager
def suppress(logger: Logger):
    try:
        yield
    except Exception as e:
        logger.exception(f"Suppressed Exception {e}")
    finally:
        pass


def open_in_browser(content, file_type: str = ".html") -> bool:
    fd, path = tempfile.mkstemp(file_type)
    os.write(fd, content)
    os.close(fd)

    return webbrowser.open("file://" + path)



import os
import sys
import typing
import abc
import itertools
import logging
import time

import httpx

from .pipelines import Pipeline
from .items import Item
from .exceptions import Dropped
from .pool import Pool
from .reporter import Reporter
from . import utils

pwd = os.getcwd()
if os.path.exists(os.path.join(pwd, "settings.py")):
    sys.path.append(pwd)
    import settings
else:
    from . import _settings_example as settings

__all__ = ["Ant", "CliAnt"]


class Ant(abc.ABC):
    response_pipelines: typing.List[Pipeline] = []
    request_pipelines: typing.List[Pipeline] = []
    item_pipelines: typing.List[Pipeline] = []

    def __init__(self):
        self._start_time = time.time()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.client = httpx.AsyncClient(**settings.HTTPX_CONFIG)
        self.pool = Pool(**settings.POOL_CONFIG)
        self.reporter = Reporter(**settings.REPORTER)

    @property
    def name(self):
        return self.__class__.__name__

    async def request(
        self,
        url: str,
        method: str = "get",
        data: httpx._models.RequestData = None,
        files: httpx._models.RequestFiles = None,
        json: typing.Any = None,
        params: httpx._models.QueryParamTypes = None,
        headers: httpx._models.HeaderTypes = None,
        cookies: httpx._models.CookieTypes = None,
        auth: httpx._auth.Auth = None,
        stream: bool = False,
    ) -> httpx.Response:
        request: httpx.Request = self.client.build_request(
            method,
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            data=data,
            files=files,
            json=json,
        )
        request = await self._pipe(request, self.request_pipelines)
        self.reporter.report(request)

        response = await utils.retry(settings.HTTP_RETRIES, settings.HTTP_RETRY_DELAY)(
            self.client.send
        )(request, auth=auth, stream=stream)

        response = await self._pipe(response, self.response_pipelines)
        self.reporter.report(response)

        return response

    async def collect(self, item: Item):
        self.logger.debug("Collect item: " + str(item))
        await self._pipe(item, self.item_pipelines)
        self.reporter.report(item)

    async def open(self):
        self.logger.info("Opening")
        for pipeline in itertools.chain(
            self.item_pipelines, self.response_pipelines, self.request_pipelines
        ):
            await utils.run_cor_func(pipeline.on_spider_open)

    async def close(self):
        await self.pool.wait_close()

        for pipeline in itertools.chain(
            self.item_pipelines, self.response_pipelines, self.request_pipelines
        ):
            await utils.run_cor_func(pipeline.on_spider_close)

        await self.client.aclose()

        self.reporter.close()

        self.logger.info("Closed")

    @abc.abstractmethod
    async def run(self):
        """App custom entrance"""

    async def main(self):
        with utils.suppress(self.logger):
            await self.open()
            await self.run()
        with utils.suppress(self.logger):
            await self.close()
        self.logger.info(
            "Run {:s} in {:f} seconds".format(
                self.__class__.__name__, time.time() - self._start_time
            )
        )

    async def _pipe(
        self,
        obj: typing.Union[Item, httpx.Request, httpx.Response],
        pipelines: typing.List[Pipeline],
    ) -> typing.Any:
        self.logger.debug("Process obj: " + str(obj))
        raw_obj = obj
        for pipeline in pipelines:
            try:
                obj = await utils.run_cor_func(pipeline.process, obj)
            except Exception as e:
                if isinstance(e, Dropped):
                    self.reporter.report(raw_obj, dropped=True)
                raise e
        return obj


class CliAnt(Ant):
    """As a http client"""

    async def run(self):
        pass



"""
Background coroutine pool with limit
"""
import typing
import asyncio
from asyncio.queues import Queue, QueueEmpty
import logging
from itertools import islice


__all__ = ["Pool"]


class Pool:
    def __init__(self, limit: int = 50):
        self._loop = asyncio.get_event_loop()
        self._limit = limit
        self.logger = logging.getLogger(self.__class__.__name__)
        self._pending_queue: Queue = Queue()
        self._done_queue: Queue = Queue()
        self._running_count = 0
        self._closed = False

    def spawn(self, coroutine: typing.Awaitable):
        """Like "asyncio.ensure_future", with concurrent limit

        Call "self.done" make sure all coroutine has been
        done.
        """
        if self._closed:
            self.logger.warning("This ant has be closed!")
            return

        if self._limit == -1 or self._running_count < self._limit:
            self._running_count += 1
            asyncio.ensure_future(coroutine, loop=self._loop).add_done_callback(
                self._done_callback
            )
        else:
            self._pending_queue.put_nowait(coroutine)

    @property
    def closed(self):
        return self._closed

    @property
    def done(self):
        return not (
            self._running_count
            or self._done_queue.qsize()
            or self._pending_queue.qsize()
        )

    async def wait_done(self):
        """Wait all coroutine to be done"""
        while not self.done:
            await self._done_queue.get()

    def close(self):
        self._closed = True

    async def wait_close(self):
        """Graceful close"""
        await self.wait_done()
        self.close()
        await self.wait_done()  # wait again avoid concurrent competition

    def as_completed(
        self,
        coros: typing.Iterable[typing.Awaitable],
        limit: int = 50,
    ) -> typing.Generator[typing.Awaitable, None, None]:
        """Like "asyncio.as_completed",
        run and iter coros out of pool.

        :param limit: set to "settings.JOB_LIMIT" by default,
        this "limit" is not shared with pool`s limit
        """
        coros = iter(coros)
        queue: Queue = Queue()
        todo: typing.List[asyncio.Future] = []

        def _done_callback(f):
            queue.put_nowait(f)
            todo.remove(f)
            try:
                nf = asyncio.ensure_future(next(coros))
                nf.add_done_callback(_done_callback)
                todo.append(nf)
            except StopIteration:
                pass

        async def _wait_for_one():
            return (await queue.get()).result()

        if limit <= 0:
            fs = {asyncio.ensure_future(cor, loop=self._loop) for cor in coros}
        else:
            fs = {
                asyncio.ensure_future(cor, loop=self._loop)
                for cor in islice(coros, 0, limit)
            }
        for f in fs:
            f.add_done_callback(_done_callback)
            todo.append(f)

        while len(todo) > 0 or queue.qsize() > 0:
            yield _wait_for_one()

    async def as_completed_with_async(
        self,
        coros: typing.Iterable[typing.Awaitable],
        limit: int = 50,
        raise_exception: bool = True,
    ) -> typing.AsyncGenerator[typing.Any, None]:
        """as_completed`s async version, can catch and log exception inside."""
        for coro in self.as_completed(coros, limit=limit):
            try:
                yield await coro
            except Exception as e:
                if raise_exception:
                    raise e
                else:
                    self.logger.exception(
                        "Get exception {:s} in "
                        '"as_completed_with_async"'.format(str(e))
                    )

    def _done_callback(self, f):
        self._running_count -= 1
        self._done_queue.put_nowait(f)
        try:
            if not self.closed and (
                self._limit == -1 or self._running_count < self._limit
            ):
                next_coroutine = self._pending_queue.get_nowait()
                self._running_count += 1
                asyncio.ensure_future(
                    next_coroutine, loop=self._loop
                ).add_done_callback(self._done_callback)
        except QueueEmpty:
            pass



"""Provide Ant`s Item and Extractor."""
import typing
from collections.abc import MutableMapping

import httpx

from .exceptions import ItemGetValueError


class CustomNoneType:
    """Different with "None" obj ("null" in json)"""

    pass


Item = typing.TypeVar("Item")


def set_value(item: Item, key: str, value: typing.Any):
    if isinstance(item, MutableMapping):
        item[key] = value
    else:
        setattr(item, key, value)


def get_value(item: Item, key: str) -> typing.Any:
    try:
        if isinstance(item, MutableMapping):
            return item[key]
        else:
            return getattr(item, key)
    except (KeyError, AttributeError) as e:
        raise ItemGetValueError from e


class Extractor:
    def __init__(self, item_cls: typing.Type[Item]):
        self.item_cls = item_cls
        self.extractors: typing.Dict[
            str, typing.Callable[[typing.Any], typing.Any]
        ] = dict()

    def add_extractor(
        self, key: str, extractor: typing.Callable[[typing.Any], typing.Any]
    ):
        self.extractors[key] = extractor

    def extract(self, res: httpx.Response) -> Item:
        item = self.item_cls()
        for key, extractor in self.extractors.items():
            set_value(item, key, extractor(res))

        return item


class NestExtractor(Extractor):
    def __init__(
        self,
        item_class: typing.Type[Item],
        root_extractor: typing.Callable[[httpx.Response], typing.Sequence],
    ):
        self.root_extractor = root_extractor
        super().__init__(item_class)

    def extract_items(self, res: httpx.Response) -> typing.Generator[Item, None, None]:
        for node in self.root_extractor(res):
            yield super().extract(node)


__all__ = [
    "Item",
    "Extractor",
    "NestExtractor",
    "get_value",
    "set_value",
]



import typing
from collections import defaultdict
import logging
import asyncio


class Record:
    def __init__(self):
        self.last_count = 0  # last report slot count
        self.count = 0
        self.dropped_last_count = 0
        self.dropped_count = 0

    def add(self, dropped: bool):
        if dropped:
            self.dropped_count += 1
        else:
            self.count += 1


class Reporter:
    def __init__(self, slot: float = 60):
        self._records: typing.DefaultDict[str, Record] = defaultdict(Record)
        self._slot = slot  # report once after one minute by default
        self._log_task = asyncio.ensure_future(self._log())
        self.logger = logging.getLogger(self.__class__.__name__)

    def report(self, obj: typing.Any, dropped: bool = False):
        self._records[obj.__class__.__name__].add(dropped)

    def close(self):
        self._log_task.cancel()
        for name, record in self._records.items():
            self.logger.warning(f"Get {record.count} {name} in total")
            self.logger.warning(f"Drop {record.dropped_count} {name} in total")

    async def _log(self):
        while True:
            await asyncio.sleep(self._slot)
            for name, record in self._records.items():
                count = record.count - record.last_count
                dropped_count = record.dropped_count - record.last_count
                record.last_count = record.count
                record.dropped_last_count = record.dropped_count
                self.logger.info(
                    f"Get {record.count} {name} in total with {count}/{self._slot} rate"
                )
                self.logger.info(
                    f"Drop {record.dropped_count} {name} in total with {dropped_count}/{self._slot} rate"
                )



"""setting module for your project"""
import os
import logging
import asyncio

import httpx

from ant_nest.exceptions import ExceptionFilter

# your ant`s class modules or packages
ANT_PACKAGES = ["ants"]
ANT_ENV = os.getenv("ANT_ENV", "development")


# httpx config, see httpx.Client.__init__ for more detail
HTTPX_CONFIG = {
    "timeout": 5.0,
    "max_redirects": 20,
    "limits": httpx.Limits(max_connections=100, max_keepalive_connections=20),
    "trust_env": True,
    "proxies": None,
    "auth": None,
    "headers": None,
    "cookies": None,
}


POOL_CONFIG = {
    "limit": 100,
}
REPORTER = {
    "slot": 60,
}


# ANT config
HTTP_RETRIES = 0
HTTP_RETRY_DELAY = 0.1


if ANT_ENV in ("development", "testing"):
    logging.basicConfig(level=logging.DEBUG)
    asyncio.get_event_loop().set_debug(True)
else:
    logging.basicConfig(level=logging.INFO)
    logging.getLogger().addFilter(ExceptionFilter())


# custom setting, eg:
# MYSQL_HOST = '127.0.0.1'
